let make_arr dim_x dim_y dim_z =
  let arr = Array.make dim_z (Array.make 0 (Array.make 0 ' ')) in
  for i = 0 to dim_z - 1 do
    arr.(i) <- Array.make dim_x (Array.make 0 ' ');
    for j = 0 to dim_x - 1 do
      arr.(i).(j) <- Array.make dim_y ' ';
    done
  done;
  arr

let array3_dim arr =
  (Array.length arr, Array.length arr.(0), Array.length arr.(0).(0))

let is_on_grid arr (ply, col, row) =
  let (plies, cols, rows) = array3_dim arr in
  not (row < 0 || row >= rows || col < 0 || col >= cols || ply < 0 || ply >= plies)
  
let is_cell arr (ply, row, col) =
  if not @@ is_on_grid arr (ply, row, col) then false
  else arr.(ply).(row).(col) != ' '

let set arr (ply, row, col) char =
  if is_on_grid arr (ply, row, col) then
    arr.(ply).(row).(col) <- char

let ( ++ ) (a, b, c) (x, y, z) = (a + x, b + y, c + z)

let all_deltas = [(0, 0, 1); (0, 0, -1); (0, 1, 0); (0, -1, 0); (1, 0, 0); (-1, 0, 0)]

let all_neighbors arr pos =
  List.filter (is_on_grid arr) (List.map (fun x -> x ++ pos) all_deltas)
               
let empty_neighbors arr pos =
  let empty p = not @@ is_cell arr p in
  List.filter empty (all_neighbors arr pos)

let random_empty_neighbor arr pos =
  let candidates = empty_neighbors arr pos in
  let len = List.length candidates in
  if len = 0 then None
  else
    let r = Random.int len in
    Some (List.nth candidates r)
  
let random_neighbor arr pos =
  let ls = all_neighbors arr pos in
  let r = Random.int (List.length ls) in
  List.nth ls r

let rec draw_path arr pos counter =
  if counter > 0 then
    begin
      let new_pos =
        match random_empty_neighbor arr pos with
        | None -> random_neighbor arr pos
        | Some v -> v in
      set arr new_pos '+';
      draw_path arr new_pos (counter - 1)
    end

let main () =
  Random.init 16;
  let arr = make_arr 20 30 3 in
  draw_path arr (0, 10, 0) 800;
  let stream = open_out "voxelpvp4.vox" in
  output_string stream "#floor_tex shared_tech/floortile1b 0.25 0.25 0 0
#wall_tex shared_tech/floortile1c 0.25 0.25 0 0
#sky_tex voxel/sky 1 1 0 0
#lamp_tex voxel/light2_white 0.5 0.5 0 0
#cell_dim 192 192 192
";
  let (plies, rows, cols) = array3_dim arr in
  for ply = 0 to plies - 1 do
    output_string stream "#ply\n";
    for row = 0 to rows - 1 do
      for col = 0 to cols - 1 do
        output_char stream arr.(ply).(row).(col)
      done;
      output_char stream '\n';
    done;
  done;
  close_out stream

let () = main ()
